#############################################################################
##
#W getroots.gi              Package Coclass                      Bettina Eick
##
## Check the periodicity of G(2,r)
##

GetRoots := function( grps )
    local p, roots, cls, G, K, H, i, j, r, c, L, M, t, k, done;

    # set up
    p := grps[1]!.prime;
    roots := [];
    cls := [];
 
    # loop
    i := 1;
    for i in [1..Length(grps)] do
        Print("go for group ",i,"\n");

        # set up
        G := grps[i];
        K := PcpGroupToFpGroup(G);
        r := CoclassSpaceGroupPC(G);

        # loop
        c := 2;
        while not IsBound(roots[i]) do
            H := PqStandardPresentation( K : Prime := p, ClassBound := c );
            M := PcGroupFpGroup(H);
            if CoclassPGroup(M) = r then
                roots[i] := ListPresentation(H);
                cls[i] := c;
            else
                c := c+1;
            fi;
        od;
    od;

    # eliminate collisions
    done := false;
    while not done do
        done := true;
        for i in [1..Length(grps)] do
            j := Filtered([1..Length(grps)], x -> roots[x] = roots[i]);
            if Length(j) > 1 then
                for k in j do
                    K := PcpGroupToFpGroup(grps[k]);
                    c := cls[k]+1;
                    H := PqStandardPresentation(K:Prime:=p,ClassBound:=c);
                    roots[k] := ListPresentation(H);
                    cls[k] := c;
                    done := false;
                od;
            fi;
        od;
    od;

    return cls;
end;

CoclassTwigInfo := function( H, K, r )
    local grps, twig, i, j, next, d, des, cap, aut;

    grps := [H];
    twig := [[Length(Factors(Size(H)))]];
    i := 1;

    while Length(grps) > 0 do
        Print("      next level with ",Length(grps)," groups \n");
        i := i + 1;
        next := [];
        twig[i] := [];
        for d in grps do
            des := PqDescendants( d : StepSize := 1 );
            des := Filtered(des, x -> CoclassPGroup(x) = r );
            cap := Filtered( des, IsCapable );

            # cut off the main line in the first step
            if i = 2 then 
                aut := List( cap, x -> false );
                cap := EliminateMainLine( K, cap, aut ); 
                des := Concatenation(cap, Filtered(des, IsTerminal));
            fi;

            # add stuff
            Append(twig[i], List(des, x -> Length(Factors(Size(x)))));
            Append(next, cap);

        od;
        Sort(twig[i]);

        # reset for next level
        grps := next;
    od;

    return twig;
end;

CoclassBranchInfo := function( G, h )
    local p, r, T, c, t, d, F, twigs, i, H, K, j, k, f, bl, bf;

    # set up
    p := G!.prime;
    r := CoclassSpaceGroupPC(G);

    # get T, d and c
    T := G;
    c := 1;
    while not IsFreeAbelian(T) do
        T := CommutatorSubgroup(T, G); Cgs(T); c := c+1;
    od;
    t := Cgs(T);
    T := SubgroupByIgs(G, t);
    d := Length(t);

    # set up for iteration
    F := PcpGroupToFpGroup( G );
    twigs := [];

    # do the initial d twigs
    for i in [1..2*d] do
        Print("    compute twig ",h+i-1,"\n");
        H := Pq( F : Prime := p, ClassBound := h+i-1 );
        K := Pq( F : Prime := p, ClassBound := h+i );
        twigs[i] := CoclassTwigInfo( H, K, r );
    od;

    # add twigs until periodicity sets in
    j := 0;
    while not twigs{[1..d]}+d = twigs{[d+1..2*d]} do
        j := j + 1;

        # eliminate 1st twig 
        twigs := twigs{[2..2*d]};

        # construct new twig
        Print("    compute twig ",h+2*d+j-1,"\n");
        H := Pq( F : Prime := p, ClassBound := h+2*d+j-1 );
        K := Pq( F : Prime := p, ClassBound := h+2*d+j );
        twigs[2*d] := CoclassTwigInfo( H, K, r );
    od;

    # get twiglength 
    k := Maximum( List( twigs, x -> Length(x)));

    # get periodicity root
    f := h+j;

    # compute bounds
    bl := Maximum( [c+p^r, c+Int(3*(k+d)/2)+1, h] );
    bf := (2*d+1)*bl + (2*r-1)*d;

    # collect all information
    return rec( f := f, bl := bl, bf := bf );
end;

