
MakeAuto := function( C, gens, imgs )
    local a;
    a := GroupHomomorphismByImagesNC( C, C, gens, imgs );
    SetKernelOfMultiplicativeGeneralMapping( a, TrivialSubgroup(C) );
    SetIsInjective( a, true );
    SetIsSurjective( a, true );
    SetIsBijective( a, true );
    return a;
end;

InducedAutoFactor := function( hom, auto )
    local gens, imgs; 
    gens := auto!.MappingGeneratorsImages;
    imgs := gens[2];
    gens := gens[1];
    imgs := List(imgs, x -> Image(hom,x));
    gens := List(gens, x -> Image(hom,x));
    return MakeAuto(Range(hom), gens, imgs);
end;

InduceAutoSK := function( auto, d )
    local l, C, S, gens, imgs, Q, genq, imgq;

    # setup
    C := Source(auto);
    l := Length(Igs(C));
    S := SubgroupByIgs( C, Igs(C){[d+1..l]} );

    # extract infos
    gens := auto!.MappingGeneratorsImages;
    imgs := ShallowCopy(gens[2]){[1..d]};

    # induce
    Q := PcpGroupByPcp( Pcp(C,S) );
    genq := Cgs(Q);
    imgq := List(imgs, x -> MappedVector( Exponents(x){[1..d]}, genq ) );
    return MakeAuto( Q, genq, imgq );
end;

ExtendAuto := function( C, auto, d )
    local l, S, gens, imgs, genc, imgc;

    # set up
    l := Length(Igs(C));
    S := SubgroupByIgs( C, Igs(C){[d+1..l]} );

    # extract infos
    gens := auto!.MappingGeneratorsImages;
    imgs := ShallowCopy(gens[2]);

    # extend
    genc := Igs(C){[1..d]};
    imgc := List( imgs, x -> MappedVector( Exponents(x), genc ) );
    gens := CgsParallel( genc, imgc );
    return MakeAuto( C, gens[1], gens[2] );
end;

InvertAuto := function( auto )
    local inds, gens, imgs, news, invs, extd;

    # check
    if IsBound(auto!.inv) then return auto!.inv; fi;

    # use induced
    inds := InduceAutoSK( auto, 8 );

    # invert
    gens := inds!.MappingGeneratorsImages;
    news := CgsParallel( gens[2], gens[1] );
    invs := MakeAuto( Source(inds), news[1], news[2] );

    # extend and return
    extd := ExtendAuto( Source(auto), invs, 8 );
    auto!.inv:= extd;
    return extd;
end;

CleanAuto := function( auto )
    local C, p, s, T, d, l, gens, imgs, i, exp, j;

    # setup
    C := Source(auto);
    p := C!.prime;
    s := C!.param;

    # get parameters
    T := Translations(C);
    d := p^(s-1)*(p-1) + Length(Pcp(C,T));
    l := Length(Pcp(C));

    # extract infos
    gens := auto!.MappingGeneratorsImages;
    imgs := ShallowCopy(gens[2]);
    gens := gens[1];

    # loop 
    for i in [1..d] do
        exp := ShallowCopy(Exponents( imgs[i] ));
        for j in [d+1..l] do exp[j] := 0; od;
        imgs[i] := MappedVector( exp, gens );
    od;

    # new auto
    return  MakeAuto(C, gens, imgs );
end;
