
#############################################################################
##
#F MatrixActionOnZ1( C, cc, aut )
##
MatrixActionOnZ1 := function ( C, cc, aut )
    local  fac, mat, con, exp, rev, add, lin, trl, coc, img, inv, l, d, j, g,
           new, i, ims;

    # get setup
    fac := C.factor;
    l := Length( fac );
    d := Length( cc.basis );

    # consider automorphism
    inv := InvertAuto( aut );
    ims := MappingGeneratorsImages(aut)[2];

    # the action on the normal subgroup as matrix
    mat := List(C.normal, x -> ExponentsByPcp(C.normal, Image(aut,x)));

    # the action on the factor group as matrix
    con := List( fac, x -> Image(inv, x) );
    exp := List( con, x -> ExponentsByPcp( fac, x ) );

    # the linear part of the action
    lin := [];
    for j  in [ 1 .. d ]  do
        coc := CutVector( cc.basis[j], l  );
        img := List( exp, x -> EvaluateCocycle( C, coc, x ) * mat );
        lin[j] := Flat( img );
    od;

    # the translation part
    coc := CutVector( cc.transl, l );
    img := List( exp, x -> EvaluateCocycle( C, coc, x ) * mat );
    img := Flat( img ) - cc.transl;

    # adjust translation 
#    rev := [];
#    for i in [1..Length(exp)] do
#        new := MappedVector( exp[i], fac );
#        new := MappedVector( Exponents(new), ims );
#        Add( rev, new );
#        Print(i, " done \n");
#    od;
    rev := List( exp, x -> Image( aut, MappedVector( x, fac ) ) );
    add := List( [ 1 .. l ], x -> fac[x]^-1 * rev[x] );
    add := Flat( List( add, x -> ExponentsByPcp( C.normal, x ) ) );

    # put the parts together
    return rec( lin := lin, trl := img + add );
end;

#############################################################################
##
#F OrbitsStabilizers( A, dom, gen, act, opr )
##
OrbitsStabilizers := function( A, dom, gen, act, opr )
    local B, b, orb, o, s;

    # set up
    if Length(dom) = 0 then return dom; fi;
    B := AbelianPcpGroup( Length(gen), List(gen, x -> 0) );
    b := Pcp(B);

    # get orbits
    orb := [];
    while Length(dom) > 0 do
        o := PcpOrbitStabilizer( dom[1], b, act, opr );
        s := List( o.stab, x -> MappedVector(Exponents(x), gen) );
        s := List( s, CleanAuto );
        Add( orb, rec( rep := dom[1],
                       len := Length(o.orbit),
                       stb := SubgroupNC(A,s) ) );
        dom := Difference( dom, o.orbit );
        #Print("      got orbit of length ",Length(o.orbit),"\n");
    od;
    return orb;
end;

#############################################################################
##
#F OrbitsInvariantComplements( cl, N, hom )
##
OrbitsInvariantComplements := function( cl, N, hom )
    local  G, H, U, C, cc, aff, tup, act, elm, fix, orb, i, com, aut, top,
           ind, per;

    # set up
    G := Source( hom );
    H := Range( hom );
    U := Image( hom, cl[1] );
    C := rec(group:=U,super:=Pcp(H,U),factor:= Pcp(U,N),normal:= Pcp(N));
    AddFieldCR(C); AddRelatorsCR(C); AddOperationCR(C); AddInversesCR(C);

    # compute one-cohomology (cobounds are trivial)
    cc := OneCocyclesEX( C );
    #Print("    cohomology has dim ",Length(cc.basis),"\n");

    # catch two trivial cases
    if IsBool(cc.transl) then return []; fi;
    if Length(cc.basis)=0 then return [ComplementCR(C, cc.transl)]; fi;

    # the affine action
    aff := function ( pt, act )
        if act = 1 then return pt; fi;
        return SolutionMat( cc.basis, pt * act.lin + act.trl );
    end;

    # the action of the group
    act := OperationOnZ1( C, cc );

    # fixed points under this action
    elm := ExponentsByRels( List( cc.basis, x -> C.char ) ) * One(C.field);
    fix := FixedPoints( elm, act, aff ); 
    #Print("    number of fixed points ",Length(fix),"\n");
    if Length( fix ) = 0  then return fix; fi;

    # consider automorphisms
    aut := GeneratorsOfGroup(cl[3]);
    ind := List( aut, x -> InducedAutoFactor( hom, x ) );
    top := List( ind, x -> MatrixActionOnZ1( C, cc, x ) );
    orb := OrbitsStabilizers( cl[3], fix, aut, top, aff );

    ## check
    per := List( top, x -> Permutation( x, fix, aff ) );
    Print(per, "\n");
    Print(List(top, x -> Order(x.lin)),"\n");
    
    # translate back
    for i in [1..Length(orb)] do
        com := IntVector( orb[i].rep * cc.basis + cc.transl );
        orb[i].rep := PreImage(hom, ComplementCR( C, com ) );
    od;
    return orb;
end;

