
#############################################################################
##
#F CoverAction2( N, pcp, mat )
##
CoverAction2 := function( N, pcp, mat )
    local d, t, r, aut, img, big;

    # set up
    d := Length(mat);
    t := Igs(N){[1..d]};

    # get automorphism
    r := List( mat, x -> MappedVector(x,t) );
    aut := GroupHomomorphismByImagesNC( N, N, t, r );

    # restrict to D
    img := List( pcp, x -> Image(aut,x) );
    big := List( img, x -> ExponentsByPcp( pcp, x) );

    # return
    return rec( mat := big, aut := aut );
end;

#############################################################################
##
#F CommutatorTails( enum, n )
##
CommutatorTails := function(enum, n)
    local res, i;
    res := [];
    for i in [1..Length(enum)] do
        if enum[i][1] > enum[i][2] and enum[i][2] > n then Add(res,i); fi;
    od;
    return res;
end;

#############################################################################
##
#F EvaluateRelator( F, pcp, e )
##
EvaluateRelator := function( F, pcp, e )
    return ExponentsByPcp( pcp, Comm(Igs(F)[e[1]], Igs(F)[e[2]] ) );
end;

#############################################################################
##
#F NiceElement( elm, base )
##
NiceElement := function( elm, base )
    local full, new, n;
    full := Concatenation( [elm], base );
    new := LLLReducedBasis( full ).basis;
    for n in new do
        if IsBool( SolutionIntMat( base, n ) ) then return n; fi;
    od;
end;

#############################################################################
##
#F ClassTwoCovers(G, all, dis)
##
ClassTwoCovers := function(arg)
    local T, d, n, l, mats, N, D, U, c, w, sub, acts, hom, CR, cc, i,
          com, prj, pre, sol, coc, fil, nul, pcp, grp, exp, e, G, dis, all;

    G := arg[1]; all := false; dis := false;
    if Length(arg) = 2 then
        all := arg[2];
    elif Length(arg) = 3 then
        all := arg[2];
        dis := arg[3];
    fi;

    # get translations
    T := Translations(G);
    d := Length(Igs(T));
    n := Length(Pcp(G,T));
    mats := LinearActionOnPcp(Pcp(G), Pcp(T));

    # get class-2 cover for T
    N := NilpotentQuotient(FreeGroup(d),2);
    D := DerivedSubgroup(N);
    l := Length(Igs(D));

    # find subspace in D
    c := GeneratorsOfGroup(Centre(Group(mats)))[1];
    w := CoverAction2( N, Pcp(D), c ).mat;
    sub := DegeneratedSubspace(w, Order(c));

    # check
    U := Subgroup(N, List(sub, x -> MappedVector(x, Igs(D)))); 
    if not IsNormal(N,U) then Error("subspace not invariant"); fi;
    if U = D then return G; fi;

    # go over to factor
    hom := NaturalHomomorphism(N,U);
    N := Image(hom);
    D := Image(hom,D); D := SubgroupByIgs(N, Cgs(D));
    pcp := Pcp(D,"snf");

    # cohomology record
    acts := List( mats, x -> CoverAction2(N, pcp, x).mat );
    CR := CRRecordByMats( G, acts );

    # H^2(G, D/U)
    Print("compute cohomology \n");
    cc := TwoCohomologyCR(CR);
    Print("  got rk(cb) = ",Length(cc.gcb),"\n");
    Print("  and cc/cb = ", cc.factor.rels,"\n");

    # find those cocycles with a prescribed commutator structure
    Print("determine cocycles with prescribed commutators \n");

    # project onto commutators
    com := CommutatorTails(CR.enumrels, n);
    prj := List( cc.gcc, x -> CutVector(x, Length(CR.enumrels)){com});

    # get special cocycle
    pre := List(CR.enumrels{com}-n, x -> EvaluateRelator(N,pcp,x));

    # find one cocycle with prescribed structure
    sol := SolutionIntMat( List(prj, Flat), Flat(pre) );
    if IsBool(sol) then 
        Print("  no solution \n");
    else
        sol := sol*cc.gcc;
        Print("  found solution \n");
    fi;

    # find all cocycles with prescribed structures
    nul := NullspaceIntMat( List(prj, Flat) ) * cc.gcc;
    nul := BaseIntMat( Concatenation( nul, cc.gcb ) );
    nul := AdditiveFactorPcp(nul, cc.gcb, 0);
    Print("  nullspace has invariants ",nul.rels,"\n");

    # construct groups
    if IsBool(sol) then return []; fi;

    if Length( nul.rels ) = 0 or not all then 
        Print("nice element and construct extension \n");
        coc := NiceElement( sol, cc.gcb );
        grp := [ExtensionCR( CR, coc )];
    else
        grp := [];
        exp := ExponentsByRels(nul.rels);
        for e in exp do
            Print("nice element and construct extension \n");
            coc := NiceElement( e*nul.prei + sol, cc.gcb );
            Add( grp, ExtensionCR( CR, coc ));
        od;

        # that's it
        if IsInt(dis) then
            Print("reduce groups \n");
            grp := DistinguishGroups( grp, G!.prime, dis );
        fi;
    fi;

    # add info
    for i in [1..Length(grp)] do
        grp[i]!.prime := G!.prime;
        grp[i]!.param := G!.param;
        grp[i]!.cocla := CoclassSpaceGroupPC(G);
    od;

    return grp;
end;

    
