#############################################################################
##
#W module.gi              Package Coclass                      Bettina Eick
##

#############################################################################
##
#F DegRatSubspace( mat, o )
##
DegRatSubspace := function( mat, o )
    local x, h, n;

    # catch the trivial case
    if Length(mat) = 1 then return IdentityMat(Length(mat)); fi;

    # get polynomial
    x := Indeterminate( Rationals );
    if IsPrime(o) then
        h := x-1;
    else
        h := (x^o - 1)/CyclotomicPolynomial( Rationals, o );
    fi;

    # find its nullspace and return
    return NullspaceMat( Value( h, mat ) );
end;

#############################################################################
##
#F AllModules(s,p)
##
AllModules := function(s,p)
    local N, G, q, d, m, lats, i, invs, mats, exts, sbsq, exsq, V, U, nath,
          unis, modus, cl, u, matu, extu, modu, emsq, facs, full;

    # set up
    N := WreathFiniteNormalizer(s,p);
    G := N!.pointgroup;
    q := p^(s-1);
    d := q*(p-1);

    # the q-th root of unity
    m := MappedVector(Exponents(Product(G!.mins)), N!.mats);

    # all lattices
    lats := [IdentityMat(d)];
    for i in [1..q-1] do
        Add( lats, NormalFormIntMat(lats[i]*(m-m^0), 2).normal );
    od;
    invs := List(lats, x -> x^-1);

    # the module of N on the lattices
    mats := [];
    exts := [];
    for i in [1..q] do

        # get the action on the i-th lattice
        mats[i] := lats[i] * N!.mats * invs[i];

        # get subspace
        emsq := AntiSymmetricPower( lats[i] * m * invs[i] );
        sbsq := DegRatSubspace( emsq, p^s );
        full := IdentityMat( Length(emsq[1]) );

        # get the action on the i-th extension
        facs := BaseSteinitzVectors(full, sbsq).factorspace;
        exsq := List( mats[i], x -> AntiSymmetricPower(x) );
        exts[i] := InducedActionFactor( exsq, facs, sbsq );
    od;

    # all point groups
    unis := UniserialSubgroups(N);

    # all actions
    modus := [];
    for cl in unis do
        u := cl.repr;
        U := SubgroupByIgs(N, u);
        for i in [0..p^cl.index-1] do
            matu := List( u, x -> MappedVector(Exponents(x), mats[i+1]));
            extu := List( u, x -> MappedVector(Exponents(x), exts[i+1]));
            modu := rec( group := U, latts := i, mats := matu, exts := extu );
            Add( modus, modu );
        od;
    od;
    return modus;
end;

#############################################################################
##
#F Characters( modu )
##
Characters := function(modu)
    local elms, exps, tops, bots;
    elms := Elements(modu.group);
    exps := List( elms, x -> ExponentsByPcp(Pcp(modu.group), x) );
    tops := List( exps, x -> TraceMat(MappedVector(x,modu.mats)));
    Print("top: ",tops,"\n");
    bots := List( exps, x -> TraceMat(MappedVector(x,modu.exts)));
    Print("sub: ",bots,"\n");
end;

#############################################################################
##
#F SpaceGroupTypes( modu )
##
SpaceGroupTypes := function( modu )
    local p, l, R, S;
    p := RelativeOrder(Igs(modu.group)[1]);
    l := Length(Igs(modu.group))+1;
    S := SplitExtensionPcpGroup( modu.group, modu.mats );
    R := SplitExtensionPcpGroup( modu.group, modu.exts );
    return DistinguishGroups( [S,R], p, l );
end;
