
#############################################################################
##
#F CentralAutsClassTwo(C)
##
CentralAutsClassTwo := function( C )
    local T, R, pcp, mat, Q, CR, cc, genQ, genR, gens, auts, 
          c, u, r, v, w, a;

    # setup 
    T := Translations(C);
    R := DerivedSubgroup(T);
    pcp := Pcp(C,R);
    mat := LinearActionOnPcp( pcp, Pcp(R) );

    # factor
    Q := PcpGroupByPcp(pcp);

    # compute one-cohom
    CR := CRRecordByMats( Q, mat );
    cc := OneCohomologyCR(CR);
    cc := NormalFormIntMat( cc.gcc, 2 ).normal;

    # translate to automorphisms
    genQ := AsList( pcp );
    genR := AsList( Pcp(R) );
    gens := Concatenation( genQ, genR );
    auts := [];
    for c in cc do
        v := CutVector( c, Length(pcp));
        w := List( [1..Length(v)], x -> genQ[x]*MappedVector(v[x], Pcp(R)));
        u := Concatenation(w,genR);
        r := CgsParallel( gens, u );
        Add( auts, MakeAuto( C, r[1], r[2] ) );
    od;
    return auts;
end;

#############################################################################
##
#F CentralAutsClassOne(C)
##
CentralAutsClassOne := function(C)
    local T, R, pcpP, pcpS, mat, P, CR, cc, genP, genS, gens, auts, c, v, w,
          u, r, a;

    # setup 
    T := Translations(C);
    R := DerivedSubgroup(T);
    pcpP := Pcp(C,T);
    pcpS := Pcp(T,R);

    # act
    mat := LinearActionOnPcp( pcpP, pcpS );
    P := PcpGroupByPcp(pcpP);
    CR := CRRecordByMats( P, mat );

    # one-cohom
    cc := OneCohomologyCR(CR);
    cc := NormalFormIntMat( cc.gcc, 2 ).normal;

    # translate to automorphisms
    genP := AsList( pcpP );
    genS := AsList( pcpS );
    gens := Concatenation( genP, genS );
    auts := [];
    for c in cc do
        v := CutVector( c, Length(pcpP));
        w := List([1..Length(v)], x -> genP[x]*MappedVector(v[x], genS));
        u := Concatenation(w,genS);
        r := CgsParallel( gens, u );
        Add( auts, MakeAuto( C, r[1], r[2] ) );
    od;
    return auts;
end;

############################################################################
##
#F LiftNormToAut(C, mat)
##
LiftNormToAut := function(C, mat)
    local T, R, pcpP, pcpS, act, new, try, j, genP, genS, gens, imsP, imsS,
          imgs, r;

    # setup 
    T := Translations(C);
    R := DerivedSubgroup(T);
    pcpP := Pcp(C,T);
    pcpS := Pcp(T,R);

    # construct action on C/T
    act := LinearActionOnPcp( pcpP, pcpS )[1];
    new := act^mat;
    try := act;
    j := 1;
    while try <> new do
        j := j+1;
        try := try * act;
    od;
    
    # set up automorphism
    genP := pcpP{[1]};
    genS := AsList(pcpS);
    gens := Concatenation(genP, genS); 
   
    imsP := [pcpP[1]^j];
    imsS := List( mat, x -> MappedVector( x, genS ) );
    imgs := Concatenation(imsP, imsS);

    r := CgsParallel( gens, imgs );
    return MakeAuto( C, r[1], r[2] );
end;

#############################################################################
##
#F NormalizerAuts(C)
##
NormalizerAuts := function(C)
    local s, p, W, U, G, V, matN, matV, mats;
    s := C!.param;
    p := C!.prime;
    W := WreathFiniteNormalizer(s,p);
    U := WreathFiniteNormalizer(W,0);
    G := U!.pointgroup;
    V := WreathInfiniteCentralizer(G);
    matN := List( Pcp(U,G), x -> MappedVector(Exponents(x), W!.mats) );
    matN := matN{[1,2]};
    matV := V!.mats;
    mats := Concatenation(matN, matV); 
    return List(mats, x -> LiftNormToAut(C,x));
end;

#############################################################################
##
#F DrawClassTwoBranchAut( G, Aut, l )
##
DrawClassTwoBranchAut := function( G, Aut, l )
    local p, s, d, t, r, n, U, V, W, title, Sheet, grp, new, tmp, pcp, 
          i, j, cl, CR, g, v, hom;

    # set up
    p := G!.prime;
    s := G!.param;
    d := p^(s-1)*(p-1);
    t := Filtered(Igs(G), x -> RelativeOrder(x) = 0);
    r := t{[d+1..Length(t)]};
    n := Length(Igs(G)) - Length(t);

    # jump to some point in T/T' 
    U := SubgroupByIgs(G, t);
    G!.isNormal := true;
    while Index(G,U) < p^l do
        U := CommutatorSubgroup(G,U);
        U := SubgroupByIgsAndIgs( G, Igs(U), r );
    od;

    # set up graphic sheet
    title := Concatenation( "Branch for |G| = ", String(p), "^", String(n));
    Sheet := GraphicSheet( title, 600, 700 );
    PrintScaleSheetClassTwo( Sheet, p, l, 25 );
    Disc( Sheet, 100, 70, 3 );

    # set up for iteration
    grp := [[U,[100, 70],Aut]];
    V := SubgroupByIgs(G, r);
    i := 1;

    # loop
    while Length(grp) > 0 do
        j := 1;
        Print("layer ",i," with ",Length(grp)," groups to go \n");
        W := CommutatorSubgroup(G, V);
        hom := NaturalHomomorphism( G, W );
        new := [];
        for cl in grp do

            # get complements
            tmp := OrbitsInvariantComplements( cl, Image(hom,V), hom );
            Print("    determined ",Length(tmp)," complements \n");

            # reduce by coclass
            if i = 1 then 
                tmp := Filtered(tmp, x -> CoclassPGroup(G/x.rep)=G!.cocla);
            fi;

            # draw groups
            for g in tmp do
                repeat
                    v := PrintVertexTreeSheet( Sheet, i, j, g.len, cl[2] );
                    j := j + 1;
                until not IsBool(v);
                Add( new, [g.rep, v, g.stb ] );
            od;
        od;

        # set up for next layer
        grp := new;
        V := W;
        i := i + 1;
    od;
end;


