#############################################################################
##
#W special.gi                 Coclass package                   Bettina Eick
##

SpecialPointGroups := function( r, p )
    local q, N, G, l, a, A, m, M, U, s, i, B, t;

    q := p^(r-1);
    N := WreathFiniteNormalizer(r,p);
    G := N!.pointgroup;
    l := Length(Igs(G));

    # the base group
    a := [Igs(G){[2..l]}];
    for i in [1..r-2] do

        # split each entry in a in p blocks of length t
        t := Length(a[1])/p; 
        a := List( a, x -> List( [1..p], y -> x{[t*(y-1)+1 .. t*y]}));

        # cut off the first entry in each block
        a := List( Concatenation(a), x -> x{[2..t]} );
    od;
    A := SubgroupByIgs(G, Flat(a));

    # the cyclic group
    m := Product(G!.mins);
    M := Subgroup(G, [m]);

    # the series
    U := Subgroup(G, Concatenation([m], Igs(A)));
    s := [U];
    for i in [1..q-1] do
        A := CommutatorSubgroup(A,M);
        U := Subgroup(N, Concatenation([m], Igs(A)));
        Add( s, U );
    od;

    # add matrices
    for i in [1..q] do
        s[i]!.mats := List(Igs(s[i]),x->MappedVector(Exponents(x),N!.mats));
        s[i]!.prime := p;
        s[i]!.param := r;
    od;

    return s;
end;

ActionOfSpecialGroup := function( U, l )
    local p, r, N, G, A, C, B, full, fixa, u, a, i, w;

    # set up
    p := U!.prime;
    r := U!.param;
    N := Parent(Parent(U));
    G := N!.pointgroup;
    A := ActionRecord(N);
    C := WreathInfiniteCentralizer(G);

    # get action
    AddActions( A, N, C );

    # get u
    u := Product(G!.mins);

    # get a
    a := [Igs(U)[r]];
    for i in [2..p^(r-1)] do a[i] := Comm(a[i-1], u); od;
    #a := Igs(U){[r..Length(Igs(U))]};

    # adjust
    w := Concatenation( [u], a );
    full := List(w,x->MappedVector(Exponents(x),A.conj[l]));
    fixa := List(w,x->MappedVector(Exponents(Image(A.nath, x)),A.acts[l]));
    return rec( igs := w, nath := A.nath, full := full, fixa := fixa );
end;

CohomologyOfSpecialGroup := function( U )
    local p, r, N, L, G, A, l, Ni, Gi, Ci, cl, cc;

    # set up
    p := U!.prime;
    r := U!.param;

    N := Parent(Parent(U));
    G := N!.pointgroup;
    A := ActionRecord(N);

    # get normalizer
    L := Normalizer(N,U);

    # get centre
    l := r - Length(Factors(Size(Centre(U))));

    # adjust
    Ni := WreathFiniteNormalizer( N, l );
    Gi := Ni!.pointgroup;
    Ci := WreathInfiniteCentralizer(Gi);
    AddActions( A, Ni, Ci );

    # set up class
    cl := rec( repr := Igs(U), 
               norm := Igs(L),
               index := l );

    # cohomology
    cc := CohomologiesOfPointGroup( N, A, cl, r );

    if TRANSLATE then 
        return List(Flat(cc), x -> ExtensionOfPointGroup(N,A,x));
    else
        return Flat(cc);
    fi;

end;

