#############################################################################
##
#W grpext.gi                   Package Coclass                   Bettina Eick
##

#############################################################################
##
#F MatrixRepOfExtension( N, A, cl )
##
MatrixRepOfExtension := function( N, A, cl )
    local k, c, d, l, mats, s, i, M, t, lat, S; 

    # set up
    k := cl.latnr;
    c := cl.cocyc;
    d := Length( A.invs[k] );
    l := Length( cl.point );

    # the point operation
    mats := List( cl.point, x -> PcpToMat( N, A.conj[k], x ) );

    # the cocycle
    if IsBool( c ) then 
        c := NullMat( l, d );
    else
        c := List( c, x -> x * A.invs[k] );
    fi;

    # enlarge point group gens
    s := [];
    for i in [1..l] do
        M := List( mats[i], x -> Concatenation( x, [0] ) );
        t := Concatenation( c[i], [1] );
        s[i] := Concatenation( M, [t] );
    od;

    # add translation subgroup
    lat := IdentityMat( d );
    for i in [1..d] do
        t := MutableIdentityMat( d+1 );
        t[d+1]{[1..d]} := lat[i];
        s[l+i] := t;
    od;

    # return group
    S := Group( s, s[1]^0 );
    S!.cocyc := c;
    return S;
end;

#############################################################################
##
#F ExtensionOfPointGroup( N, A, cl )
##
ExtensionOfPointGroup := function( N, A, cl )
    local S, l, h, k, p, c, n, d, m, s, t, z, coll, i, e, u, v, o, H, j;

    # the matrix group
    S := MatrixRepOfExtension( N, A, cl );
    l := GeneratorsOfGroup( S );

    # general set up
    h := Pcp( SubgroupByIgs( N, cl.point ) );
    k := cl.latnr;
    p := N!.prime;
    c := S!.cocyc;

    # dimensions
    n := Length(h);
    d := Length(l) - n;

    # matrices etc.
    m := l{[1..n]};
    s := List( m, x -> x ^ -1 );
    t := l{[n+1..n+d]};
    z := List( [1..n], x -> 0 );

    # set up 
    coll := FromTheLeftCollector( n + d );
    
    # and add in relations
    for i in [1..n] do
        SetRelativeOrder( coll, i, p );

        # power
        e := ExponentsByPcp( h, h[i]^p );
        u := s[i]^p * MyMappedVector( e, m, s );
        v := u[d+1]{[1..d]};
        o := ObjByExponents( coll, Concatenation( e, -v ) );
        SetPower( coll, i, o );

        # top conjugates
        for j in [1..i-1] do
            e := ExponentsByPcp( h, h[i]^h[j] );
            u := s[j] * s[i] * m[j] * MyMappedVector( e, m, s );
            v := u[d+1]{[1..d]};
            o := ObjByExponents( coll, Concatenation( e, -v ) );
            SetConjugate( coll, i, j, o );
        od;

        # bottom conjugates
        for j in [1..d] do
            u := s[i] * t[j] * m[i];
            v := u[d+1]{[1..d]};
            o := ObjByExponents( coll, Concatenation( z, v ) );
            SetConjugate( coll, n+j, i, o );
        od;
    od;
    
    # construct group
    UpdatePolycyclicCollector( coll );
    H := PcpGroupByCollectorNC( coll );

    # add info and return
    H!.mats := l;
    H!.prime := N!.prime;
    H!.param := N!.param;
    H!.cclas := cl!.cclas;
    return H;
end;

#############################################################################
##
#F FpExtensionOfPointGroup( N, A, cl )
##
FpExtensionOfPointGroup := function( N, A, cl )
    local S, l, h, k, p, c, n, d, m, s, t, z, F, f, r, i, e, u, v, w, a, H, j;

    # the matrix group
    S := MatrixRepOfExtension( N, A, cl );
    l := GeneratorsOfGroup( S );

    # general set up
    k := cl.latnr;
    p := N!.prime;
    c := S!.cocyc;

    # dimensions
    n := Length(cl.point);
    d := Length(l) - n;

    # matrices etc.
    m := l{[1..n]};
    s := List( m, x -> x ^ -1 );
    t := l{[n+1..n+d]};
    z := List( [1..n], x -> 0 );

    # set up 
    F := FreeGroup( n + d );
    f := GeneratorsOfGroup( F );
    r := [];
    
    # and add in relations
    for i in [1..n] do

        # power
        e := ExponentsByPcp( h, h[i]^p );
        u := s[i]^p * MyMappedVector( e, m, s );
        v := u[d+1]{[1..d]};
        w := Concatenation( e, -v );
        Add( r, f[i]^p / MappedVector( w, f ) );

        # top conjugates
        for j in [1..i-1] do
            e := ExponentsByPcp( h, h[i]^h[j] );
            u := s[j] * s[i] * m[j] * MyMappedVector( e, m, s );
            v := u[d+1]{[1..d]};
            w := Concatenation( e, -v );
            a := f[i]^f[j] / MappedVector( w, f );
            if a <> One( F ) then Add( r, a ); fi;
        od;

        # bottom conjugates
        for j in [1..d] do
            u := s[i] * t[j] * m[i];
            v := u[d+1]{[1..d]};
            w := Concatenation( z, v );
            a := f[n+j]^f[i] / MappedVector( w, f );
            if a <> One( F ) then Add( r, a ); fi;
        od;
    od;

    for i in [1..d] do
        for j in [1..i-1] do
            Add( r, Comm( f[n+i], f[n+j] ) );
        od;
    od;
    
    H := F / r;
    H!.cclas := cl!.cclas;
    return H;
end;

