#############################################################################
##
#W  system.gi                   Package Coclass                  Bettina Eick
##

#############################################################################
##
#F ConjugatedWord( a, b )
##
ConjugatedWord := function( a, b )
    local c;
    c := Reversed( b );
    c := List( c, x -> [x[1], -x[2]] );
    return Concatenation( c, a, b );
end;

#############################################################################
##
#F PoweredWord( a, e )
##
PoweredWord := function( a, e )
    local c;
    if e = 0 then return []; fi;
    if e < 0 then 
        c := Reversed( a );
        c := List( c, x -> [x[1], -x[2]] );
        e := -e;
    else
        c := StructuralCopy( a );
    fi;
    return Concatenation( List( [1..e], x -> c ) );
end;

#############################################################################
##
#F CompactedWord( w )
##
CompactedWord := function( w )
    local v, todo, i, j;

    # the trivial case
    if Length( w ) = 0 then return w; fi;
    v := List( w, x -> StructuralCopy(x) );

    # reduce v
    todo := true;
    while todo do
        todo := false;
        i := 1;
        j := 2;
        while j <= Length( v ) do
            if v[i][1] = v[j][1] then
                v[i][2] := v[i][2] + v[j][2];
                v[j][2] := 0;
                j := j + 1;
                todo := true;
            else
                i := j;
                j := i + 1;
            fi;
        od;
        for i in [1..Length(v)] do
            if v[i][2] = 0 then 
                Unbind(v[i]);
                todo := true;
            fi;
        od;
        v := Compacted(v);
    od;
    return v;
end;

#############################################################################
##
#F AddSystemFFE( sys, vec )
##
AddSystemFFE := function( sys, vec )
    local  d, j, cv, i;

    # shift vec trough sys
    repeat
        d := DepthVector( vec );
        j := Position( sys.deps, d );
        if not IsBool( j )  then
            cv := vec[d];
            AddRowVector( vec, sys.base[j], - cv );
        fi;
    until IsBool( j );
    if d > Length( vec )  then return; fi;

    # enlarge system 
    vec := vec[d]^-1 * vec;
    sys.blen := sys.blen + 1;
    sys.base[sys.blen] := vec;
    sys.deps[sys.blen] := d;

    # clear entries
    for i in [1..sys.blen-1] do
        if sys.base[i][d] <> 0 * sys.base[i][d] then 
            cv := sys.base[i][d];
            AddRowVector( sys.base[i], vec, - cv );
        fi;
    od;
end;

#############################################################################
##
#F AddSystemINT( sys, vec )
##
AddSystemINT := function( base, vec )
    local  dep, d, j, cv, cb, e, tmp;
    dep := List( base, DepthVector );
    repeat
        d := DepthVector( vec );
        j := Position( dep, d );
        if not IsBool( j )  then
            cv := vec[d];
            cb := base[j][d];
            if IsInt( cv / cb )  then
                AddRowVector( vec, base[j], - cv / cb );
            else
                e := Gcdex( cv, cb );
                tmp := base[j];
                base[j] := e.coeff1 * vec + e.coeff2 * tmp;
                vec := e.coeff3 * vec + e.coeff4 * tmp;
            fi;
        fi;
    until IsBool( j );
    if d <= Length( vec )  then
        vec := SignInt( vec[d] ) * vec;
        base[Length( base ) + 1] := vec;
    fi;
end;

#############################################################################
##
#F ClearSystemINT( base )
##
ClearSystemINT := function( base )
    return NormalFormIntMat( base, 2 ).normal;
end;

