
#############################################################################
##
## Diag( mat )
##
Diag := function( mat )
    local  diag, i;
    diag := [];
    i := 1;
    while i <= Length( mat ) and i <= Length( mat[1] )  do
        diag[i] := mat[i][i];
        i := i + 1;
    od;
    return diag;
end;

#############################################################################
##
## LowerCentral(G, l)
##
LowerCentral := function(G, l)
    local s, U, i;
    s := [G];
    U := G;
    for i in [1..l] do
        U := CommutatorSubgroup(G,U);
        Add(s, U);
    od;
    return s;
end;

#############################################################################
##
## ProPCohomology(G, i, fac, sys, ker, img, lat)
##
ProPCohomology := function( G, i, fac, sys, ker, img, lat)
    local t, l, S, mat, C, cci, cbi, dia, all, bat, cb, cc, cf, n, 
          b, c, d, o, k, pos, cck, cbk, cfk, h;

    # set up
    if IsInt(lat) then t := lat; else t := 0; fi;
    l := LowerCentral(G, i+t+1);
    S := l[i+1];

    if fac then
        mat := LinearActionOnPcp( Pcp(G,S), Pcp(S) );
        C := CRRecordByMats( G/S, mat );
    else
        mat := LinearActionOnPcp( Pcp(G), Pcp(S) );
        C := CRRecordByMats( G, mat );
    fi;

    if not IsInt(lat) then 
        cf := TwoCohomologyCR(C);
        Print("coh has invariants ", Collected(cf.factor.rels),"\n");
    fi;

    if sys then
        cbi := TwoCoboundariesCR( C );
        cci := IntTwoCocycleSystemCR( C );
        dia := Diag( NormalFormIntMat(TransposedMat(cci.base),1).normal ); 
        Print("find ccsys with invars ", Collected(dia),"\n");
        Print("find cbssys of rank ",RankMat(cbi),"\n");
    fi;

    if ker then
        d := Length(mat[1]);
        b := Position(List(Igs(G), RelativeOrder), 0);
        c := Filtered(C.enumrels, x -> (x[1]>x[2] and x[2]>=b));
        Append(c,Filtered(C.enumrels, x -> (x[1]<x[2] and x[2]-x[1]>=b)));
        Append(c,Filtered(C.enumrels, x -> (x[1]=x[2] and x[1]>=b)));
        c := List(c, x -> Position(C.enumrels,x)); Sort(c);
        pos := Flat(List(c, x -> [(x-1)*d+1..x*d]));
    fi;
       
         
    if IsInt(lat) then 
        if not sys then return fail; fi;
        all := List([1..t],n->List(Cgs(l[i+n+1]),x->ExponentsByPcp(Pcp(S),x)));
        bat := List( all, x -> DirectSumMat( List( [1..cci.len], y -> x) ) );
        for n in [1..t] do

            # get H^2(G/T_i, T_i/T_{i+n})
            cc := IntKernelCR( C, cci, all[n], bat[n] );
            cb := BaseIntMat( Concatenation( cbi, bat[n] ) );
            cf := AdditiveFactorPcp(cc, cb, 0);
            o := Length(Factors(Product(cf.rels)));
            Print(" lattice ",n," yields: ",o,"\n");

            # get tails to powers+commutators of T_0/T_i
            if ker then
                cck := BaseIntMat(List(cc, x -> x{pos}));
                cbk := BaseIntMat(List(cb, x -> x{pos}));
                cfk := AdditiveFactorPcp(cck,cbk,0);
                k := Length(Factors(Product(cfk.rels)));
                Print(" -- kernel ",n," yields: ",o-k,"\n");
            fi;
 
            # get image of bat[n-1]/bat[n]
            if img and n > 1 then
                cc := LatticeIntersection(cc, bat[n-1]);
                cc := BaseIntMat(Concatenation(cc, cb));
                cf := AdditiveFactorPcp(cc,cb,0);
                h := Length(Factors(Product(cf.rels)));
                Print(" -- image ",n," yields: ",h,"\n");
            fi;
        od;
    fi;
end;

#############################################################################
##
## ProPIntCohomology( G, i, t, fac, fil )
##
ProPIntCohomology := function( G, i, t, fac, fil )
    local l, S, m, C, cbi, sys, cbr, ccr, lat, bat, res, n, cc, cb, cf, 
          j, fl, qo;

    l := LowerCentral(G, i+t+1);
    S := l[i+1];

    # set up
    if fac = true then 
        m := LinearActionOnPcp( Pcp(G,S), Pcp(S) );
        C := CRRecordByMats( G/S, m );
    else
        m := LinearActionOnPcp( Pcp(G), Pcp(S) );
        C := CRRecordByMats( G, m );
    fi;

    # precompute coboundaries
    cbi := TwoCoboundariesCR( C );

    # precompute cocycle-system
    sys := IntTwoCocycleSystemCR( C );
    ccr := Diag( NormalFormIntMat(TransposedMat(sys.base),1).normal ); 

    # precompute lattices
    lat := List([1..t],n->List(Cgs(l[i+n+1]),x->ExponentsByPcp(Pcp(S), x)));

    # precompute blowing
    bat := List( lat, x -> DirectSumMat( List( [1..sys.len], y -> x) ) );

    # message
    Print("find ", Collected(ccr)," modulo ",RankMat(cbi),"\n");

    # loop
    res := [];
    for n in [1..t] do

        # cocycles
        cc := IntKernelCR( C, sys, lat[n], bat[n] );

        # coboundaries
        cb := BaseIntMat( Concatenation( cbi, bat[n] ) );

        # factor
        cf := AdditiveFactorPcp(cc, cb, 0);

        # message
        Print("  ",n," yields: ",cf.rels,"\n");

        # filtration
        if fil then 
            cf.filt := [cc];
            for j in [1..n-1] do
                fl := BaseIntersectionIntMats( cc, bat[j] );
                fl := BaseIntMat( Concatenation( fl, cbi ) );
                Add( cf.filt, fl );

                # message
                qo := AdditiveFactorPcp( fl, cb, 0 );
                Print("    ", j+1," gives ", qo.rels ,"\n");
            od;
            Add( cf.filt, cb );
        fi;

        # store
        Add( res, cf );
    od;
    return res;
end;

CocycleByExtension := function( hom )
    local F, N, f, p, rels, i, l, r, j, pcpN, M, pcpM, vec, img;

    F := Image(hom);
    N := Kernel(hom);
    f := AsList( Pcgs(F));
    p := PrimePGroup(F);

    # get rels
    rels := [];
    for i in [1..Length(f)] do
        l := PreImagesRepresentative(hom, f[i])^p;
        r := PreImagesRepresentative(hom, f[i]^p);
        Add( rels, l/r );
        for j in [i+1..Length(f)] do
            l := Comm( PreImagesRepresentative(hom, f[i]), 
                       PreImagesRepresentative(hom, f[j]));
            r := PreImagesRepresentative(hom, Comm(f[i],f[j]));
            Add( rels, l/r );
        od;
    od; 

    # rewrite abelian
    pcpN := Pcgs(N);
    M := PcGroupToPcpGroup(N);
    pcpM := Pcp(M,"snf");
    for i in [1..Length(rels)] do
        vec := ExponentsOfPcElement( pcpN, rels[i] );
        img := MappedVector( vec, Igs(M) );
        rels[i] := ExponentsByPcp( pcpM, img );
    od;
    return rels;
end;

