############################################################################# 
##
##  checkiso.gi                  Package coclass                 Bettina Eick
##
##  distinguish groups
##

#############################################################################
#############################################################################
##
# Distinguish
##

#############################################################################
##
#F SpaceGroupFactorByPrimePower( G, p, l )
##
SpaceGroupFactorByPrimePower := function( G, p, l )
    local T, t, s, U;
    t := Filtered( Igs(G), x -> RelativeOrderPcp( x ) = 0 );
    s := List( t, x -> x^(p^l) );
    U := Subgroup( G, s );
    return G/U;
end;

#############################################################################
##
#F StandardPresentation
##
SPFpGroup := function( G, p, l )
    return PqStandardPresentation( G : Prime := p, ClassBound := l);
end;

SPPcpGroup := function( G, p, l )
    return SPFpGroup( PcpGroupToFpGroup( G ), p, l );
end;

SPClass := function( G, cl, p, l )
    return SPFpGroup( FpExtensionOfPointGroup( G, cl ), p, l );
end;

#############################################################################
##
#F ListPresentation( F )
##
ListPresentation := function( F )
    local f, r, l;
    f := GeneratorsOfGroup(F);
    r := RelatorsOfFpGroup(F);
    l := [Length(f), List( r, ExtRepOfObj )];
    return l;
end;

#############################################################################
##
#F DistinguishGroups( grps, p, l ) . . . . . . . .         use finite factors
##
DistinguishGroups := function( grps, p, l )
    local i, pres, subs;
    pres := [];
    subs := [];
    for i in [1..Length(grps)] do
        pres[i] := ListPresentation( SPFpGroup( grps[i], p, l ) );
        Print("  ",i," has p-coclass ", pres[i][1] - l, "\n");
        if pres[i] in pres{[1..i-1]} then 
            Print("  -- isomorphism detected \n");
        else
            Add( subs, grps[i] );
        fi;
    od;
    return subs;
end;

#############################################################################
##
#F DistinguishClasses( G, grps, p, l ) . . . . . . . .     use finite factors
##
DistinguishClasses := function( G, grps, p, l )
    local i, pres;
    pres := [];
    for i in [1..Length(grps)] do
        pres[i] := ListPresentation( SPClass( G, grps[i], p, l ) );
        Print("  ",i," has p-coclass ", pres[i][1] - l, "\n");
        if pres[i] in pres{[1..i-1]} then 
            Print("  -- isomorphism detected \n");
        fi;
    od;
    return pres;
end;

#############################################################################
##
#F FingerprintPointGroup( G, U ) . . . . . . . . . . . . . .compute character
##
FingerprintPointGroup := function( G, U )
    local cl, mt;
    cl := Elements(U);
    cl := PcpOrbitsStabilizers( cl, Pcp(U), Pcp(U), OnPoints );
    cl := List( cl, x -> x.repr );
    mt := List( cl, x -> PcpToMat( G, G!.mats, x ) );
    mt := List( mt, TraceMat );
    Sort(mt);
    return mt;
end;

