
poly := function (j, k, u, v)
   local x, f, c;
   x := Indeterminate(Rationals);
   f := ((-1)^j * x^(2*j-5) * (x + 1)^((-j + 2) mod 9)
         * (x+2)^(j-3))
         * (x^2 + 3 * x + 3) * (x + 1)^u * (x^5 + 1)^v;
   c := CoefficientsOfUnivariatePolynomial (f);
   if Length(c) < k + j - 3 then
        c := Concatenation(c, List([Length(c)+1..k + j-3], x -> 0));
   fi;
   return c;
end;

e := function (k, i, j)
   local l;
   l := ((k - i + j - 3) / 6);
   if IsInt(l) then return l; fi;
   return Int(l) + 1;
end;

SkeletonGroupR := function (j, k, u, v)
    local F, f, a, X, T, i, S, R, t0, t, t1, lhs, rhs, coe;

    # check
    if not j >= 7 then return fail; fi;
    if not j <= k and k <= 2*j then return fail; fi;
   
    # set up
    F := FreeGroup(k+2);
    f := GeneratorsOfGroup(F);
    a := f[1]; t0 := f[2]; t := t0; t1 := f[3];
    X := f{[2..k+2]};

    T := [X[1]];
    for i in [1..k] do T[i + 1] := Comm(X[i], a); od;
    T := T{[2..k+1]};
    X := f{[3..k+2]};
    S := List([1..Length(X)], i -> X[i]/T[i]);

    R := [ a^9, (t*a^3)^3,  Comm(Comm(t,a^3),t),  
           Comm(t,t^(a^4))*Comm(t^a,t^(a^3)), 
           Comm(t,t^(a^5))*Comm(t^(a^2),t^(a^3)) ];

    lhs := Comm(t1, t0);
    coe := poly(j,k,u,v);
    rhs := Product(List( [2*j-5 .. k+j-4], i -> 
                T[i-j+3]^(-coe[i+1] mod (3^e(k,i,j)))));

    R := Concatenation(S, R, [lhs/rhs]);
    return F/R;
end;

# limit class to k+1
AllSkeletonGroupsR := function( j, k )
    local U, V, G, i;

    U := [1,1,1,1,1,1,1,1,1,3,3,3,3,3,0,0,0];
    V := [4,1,7,3,0,6,2,5,8,1,3,0,6,2,1,3,0];
    G := [];
    for i in [1..17] do
        G[i] := SkeletonGroupR(j,k,U[i],V[i]);
    od;
    return G;
end;

BranchLimitR := function(j)
    local f;
    f := j-13;
    if IsInt(j/3) then f := f+1; fi;
    return f;
end;

PrintTwigR := function(G, i,j,k)
    local a, b, file, ii, jj, kk, xx;
    ii := String(i);
    jj := String(j);
    kk := String(k);
    if k = 0 then xx := String(100); else xx := kk; fi;
    file := Concatenation("twigsR/twig_",ii,"_",jj,"_",kk,".gi");
    PrintTo(file,"Twig_",ii,"_",jj,"_",kk," := [\n");
    for a in [1..Length(G)] do
        AppendTo(file,"[");
        for b in [1..Length(G[a])] do
            AppendTo(file,"[",G[a][b][1],",",G[a][b][2],"]");
            if b < Length(G[a]) then
                AppendTo(file,",");
            else
                AppendTo(file,"]");
            fi;
        od;
        if a < Length(G) then
            AppendTo(file,",\n");
        else
            AppendTo(file,"]\n");
        fi;
    od;
    AppendTo(file,";\n\n");
    AppendTo(file,"TWIG[",ii,"][",jj,"][",xx,"]:= ");
    AppendTo(file,"Twig_",ii,"_",jj,"_",kk,";\n");
    AppendTo("summR.gi", "Add(list, [",ii,",",jj,",",kk,"]);\n");
end;

GetTwigInR := function( i,j,k )
    local l, U, V, G, bran, T, S;
    Print("GetTwigInR starts ",[i,j,k]," \n");
    l := k+12;
    U := [1,1,1,1,1,1,1,1,1,3,3,3,3,3,0,0,0];
    V := [4,1,7,3,0,6,2,5,8,1,3,0,6,2,1,3,0];
    G := SkeletonGroupR( j, j+l, U[i], V[i] );
    bran := not (k=BranchLimitR(j));
    T := CoclassTree(G, 3, j+l, bran);
    S := TwigCanoForm(T);
    PrintTwigR(S, i, j, k);
end;

