RequirePackage("anupq");
RequirePackage("sophus");
RequirePackage("laguna");
RequirePackage("modisom");

ProPMaxClass := function(p)
    local f,c;
    f := CyclotomicPolynomial(Rationals,p);
    c := CompanionMat(f);
    return SplitExtensionPcpGroup(AbelianPcpGroup([p]),[c]);
end;

IsTerminal := function( G ) return not IsCapable(G); end;

EliminateMainLine := function( K, cap, aut )
    local sub;
    sub := Filtered([1..Length(cap)], x -> aut[x] = aut[1]);
    while Length(sub) > 1 do
        if IsIsomorphicPGroup( K, cap[sub[1]] ) then
            sub := [sub[1]];
        else
            sub := sub{[2..Length(sub)]};
        fi;
    od;
    sub := Difference( [1..Length(cap)], sub );
    return cap{sub};
end;

Coclass := function(G)
    local l, n;
    l := Length(LowerCentralSeries(G))-1;
    n := Length(Factors(Size(G)));
    return n-l;
end;

TableByGroup := function(G)
    local p, A;
    p := PrimePGroup(G);
    A := GroupRing(GF(p), G);
    return TableByWeightedBasisOfRad(A);
end;

FindType := function( rep, alg, L )
    local j;
    for j in [1..Length(rep)] do
        if CompareTables(alg[rep[j][1]], L) = true then 
            return j;
        fi;
    od;
    return false;
end;

IsomorphismReps := function(grp)
    local alg, rep, can, i, j;
    alg := List(grp, x -> TableByGroup(x));
    can := List(alg, x -> CanonicalFormOfTable(x));
    rep := [[1]];
    for i in [2..Length(grp)] do
        j := FindType(rep, can, can[i]);
        if j=false then 
            Add(rep, [i]);
        else
            Add(rep[j], i);
        fi;
    od;
    return rep;
end;

#############################################################################
##
#F CoclassBranchReduced(G, p, s)
##
CoclassBranchReduced := function( G, p, s)
    local F, H, K, tree, i, j, k, d, des, cap, aut, lie, iso;

    # set up p-quotients of G
    if IsPcpGroup(G) then 
        F := PcpGroupToFpGroup(G);
    elif IsPcGroup(G) then 
        F := FpGroupPcGroup(G);
    elif IsFpGroup(G) then 
        F := G;
    fi;
    H := Pq(F : Prime := p, ClassBound := s);
    K := Pq(F : Prime := p, ClassBound := s+1); 

    # construct tree
    tree := [[[false, [H]]]]; i := 1;
    while Length(tree[i]) > 0 do
        i := i+1;
        tree[i] := [];
        for j in [1..Length(tree[i-1])] do
            Print("start alg ",j," at level ",i-1,"\n");
            for k in [1..Length(tree[i-1][j][2])] do
                d := tree[i-1][j][2][k];
                des := PqDescendants( d : ClassBound := s+i-1);
                cap := Filtered( des, IsCapable );
            
                 # cut of main line if desired
                if i = 2 then 
                    aut := List( cap, x -> Size(AutomorphismGroup(x)) );
                    SortParallel( aut, cap, function(a,b) return a>b; end );
                    cap := EliminateMainLine( K, cap, aut ); 
                fi;

                # determine isomorphisms
                iso := IsomorphismReps(cap);

                # add 
                Append(tree[i], List(iso, x -> [j, cap{x}]));

            od;
            tree[i-1][j][2] := Length(tree[i-1][j][2]);
        od;
    od;

    return tree;
end;

