#############################################################################
##
#W xmeta.gi                 Package Coclass                      Bettina Eick
##
IsMetabelian := function(G) return IsAbelian(DerivedSubgroup(G)); end;

LCSFactors := function(G, s)
    local U, l, i, V;
    U := G;
    l := [];
    for i in [1..s] do
        V := CommutatorSubgroup(U,G);
        Add(l, AbelianInvariants(U/V));
        U := V;
    od;
    return l;
end;

#############################################################################
##
#F DrawCoclassTreeMetabelian(G, p, s, l) . . draw tree or branch for G from 
##                                           class s to l
##
##  -- metabelian groups only
##
DrawCoclassTreeMetabelian := function( arg )
    local G, H, F, K, p, s, l, r, opt, Title, Sheet, 
          grp, nex, des, ter, cap, res, aut,
          i, j, d, g, v, a, c;

    # set up
    G := arg[1];
    s := arg[2];
    l := arg[3];
    if l = false then l := 100; fi;
    if Length(arg) = 4 then opt := arg[4]; else opt := rec(); fi;
    p := Factors(RelativeOrdersOfPcp(Pcp(G))[1])[1];

    # set up p-quotient of G
    if IsPcpGroup(G) then 
        F := PcpGroupToFpGroup(G);
    elif IsPcGroup(G) then 
        F := FpGroupPcGroup(G);
    elif IsFpGroup(G) then 
        F := G;
    fi;
    H := Pq(F : Prime := p, ClassBound := s);
    r := Length(Factors(Size(H)))-s;

    # if desired then determine next class
    if IsBound(opt.bran) then 
        K := Pq(F : Prime := p, ClassBound := s+1); 
    fi;

    # set up graphic sheet with scale
    Title := "coclass tree";
    Sheet := GraphicSheet( Title, SSIZE[1], SSIZE[2] );
    PrintScale( Sheet, p, s, s+r, s+25 );
    Disc( Sheet, 100, 70, 3 );

    # set up for iteration
    grp := [[H,[100, 70]]];
    res := [1];
    c := false;
    
    # compute iterated descendants
    for i in [1..l-s] do
        nex := [];
        j := 1;
        res[i+1] := 0;
        for d in grp do
            des := PqDescendants( d[1] : StepSize := 1 );
            des := Filtered( des, x -> IsMetabelian(x));
            ter := Filtered( des, IsTerminal );
            cap := Filtered( des, IsCapable );
            
            # store
            res[i+1] := res[i+1] + Length(des);

            # cut of main line if desired
            if IsBound(opt.bran) and i = 1 then
                aut := List( cap, x -> Size(AutomorphismGroup(x)) );
                SortParallel( aut, cap, function(a,b) return a>b; end );
                cap := EliminateMainLine( K, cap, aut ); 
                res[i+1] := res[i+1] - 1;
            fi;


            # draw capables
            for g in cap do
                if IsBound(opt.func) then c:=opt.func(g); fi;
                repeat
                    v := PrintVertex( Sheet, i, j, c, d[2]); 
                    j := j + 1;
                until not IsBool(v); 
                Add( nex, [g, v] );
            od;
            
            # draw terminals
            if IsBound(opt.full) or IsBound(opt.func) then
                for g in ter do
                    if IsBound(opt.func) then c:=opt.func(g); fi;
                    repeat
                        v := PrintVertex( Sheet, i, j, c, d[2]); 
                        j := j + 1;
                    until not IsBool(v); 
                od;
            else
                a := Length(ter);
                if a > 0 then 
                    repeat
                        v := PrintVertex( Sheet, i, j, a, d[2] );
                        j := j + 1;
                    until not IsBool(v);
                fi;
            fi;
        od;

        # reset for next level
        grp := nex;
        if Length(grp) = 0 then return res; fi;
    od;
    return res;
end;

#############################################################################
##
#F ProbeBranchDepth(G, s, l)
##
NextLevel := function(G, K)
    local des, aut;

    # get metabelian descendants
    des := PqDescendants( G : StepSize := 1 );
    des := Filtered( des, x -> IsMetabelian(x));
    des := Filtered( des, IsCapable );

    # sort by autos
    aut := List( des, x -> Size(AutomorphismGroup(x)) );
    SortParallel( aut, des, function(a,b) return a>b; end );

    # cut of main line if desired
    if Size(G) = Size(K) then 
        des := EliminateMainLine( K, des, aut ); 
    fi;
    if Length(des) = 0 then return fail; fi;

    # get maximal autos
    des := des{Filtered([1..Length(des)], x -> aut[x]=Maximum(aut))};

    # set up for next level
    return Random(des);
end;

ProbeBranchDepth := function( G, s, l )
    local p, F, H, K, L, res, i;

    # set up
    p := Factors(RelativeOrdersOfPcp(Pcp(G))[1])[1];

    # set up p-quotient of G
    if IsPcpGroup(G) then 
        F := PcpGroupToFpGroup(G);
    elif IsPcGroup(G) then 
        F := FpGroupPcGroup(G);
    elif IsFpGroup(G) then 
        F := G;
    fi;
    H := Pq(F : Prime := p, ClassBound := s);
    K := Pq(F : Prime := p, ClassBound := s+1); 
    res := 1;
    
    # compute iterated descendants
    for i in [1..l] do
        L := [H];
        repeat
            Add(L, NextLevel(L[Length(L)], K));
        until L[Length(L)] = fail; 
        res := Maximum( res, Length(Factors(Size(L[Length(L)-1])/Size(H))));
        Print(i," th try yields ", res, "\n");
    od;
    return res;
end;

#############################################################################
##
#F CoclassBranchMetabelian(G, p, s)
##

PrintTree := function(T)
    local Title, Sheet, i, y, j, t, s, x;

    # set up graphic sheet with scale
    Title := "coclass tree";
    Sheet := GraphicSheet( Title, SSIZE[1], SSIZE[2] );
    Disc( Sheet, 100, 70, 3 );
    T[1][1][1] := [100, 70];

    # print
    for i in [2..Length(T)] do
        y := 70+DEPTH*(i-1);
        x := 100 - WIDTH;
        for j in [1..Length(T[i])] do
            t := T[i][j];
            s := T[i-1][t[1]][1];
            x := Maximum( x + WIDTH, s[1]);
            Disc( Sheet, x, y, 3);
            Line( Sheet, s[1], s[2], x-s[1], y-s[2] );
            if t[2] <> 1 then 
                Text( Sheet, FONTS.small, x+2, y, String(t[2]) );
            fi;
            T[i][j][1] := [x,y];
        od;
    od;
end;

CoclassBranchMetabelian := function( G, p, s)
    local F, H, K, tree, i, j, d, des, ter, cap, aut, T;

    # set up p-quotient of G
    if IsPcpGroup(G) then 
        F := PcpGroupToFpGroup(G);
    elif IsPcGroup(G) then 
        F := FpGroupPcGroup(G);
    elif IsFpGroup(G) then 
        F := G;
    fi;
    H := Pq(F : Prime := p, ClassBound := s);
    K := Pq(F : Prime := p, ClassBound := s+1); 

    # construct tree
    tree := [[[false, H]]]; i := 1;
    while Length(tree[i]) > 0 do
        i := i+1;
        tree[i] := [];
        for j in [1..Length(tree[i-1])] do
            Print("start group ",j," at level ",i-1,"\n");
            d := tree[i-1][j][2];
            if IsGroup(d) then 
                des := PqDescendants( d : StepSize := 1 );
                des := Filtered( des, IsMetabelian);
                des := Filtered( des, x -> AbelianInvariants(x)=[p,p]);

                ter := Filtered( des, IsTerminal );
                cap := Filtered( des, IsCapable );
            
                 # cut of main line if desired
                if i = 2 then 
                    aut := List( cap, x -> Size(AutomorphismGroup(x)) );
                    SortParallel( aut, cap, function(a,b) return a>b; end );
                    cap := EliminateMainLine( K, cap, aut ); 
                fi;

                # add capables
                Append( tree[i], List(cap, g -> [j,g]) );

                # add terminals
                Add( tree[i], [j, Length(ter)] );

                # get rid of original group
                tree[i-1][j][2] := 1;
            fi;
        od;
    od;
    Unbind(tree[i]);
    T := TwigCanoForm(tree);
    PrintTree(T);
    return T;
end;


